--[[
LibRule - library for item select rules

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:
	
Free Software Foundation, I.,
51 Franklin Street, Fifth Floor,
Boston, MA  02110-1301, USA.
--]]

local MAJOR = 'LibRule-1.0'
local MINOR = 2
local lib = LibStub:NewLibrary( MAJOR, MINOR)
if not lib then return end -- No Upgrade needed.

local qualities = {}
local cache = {}

------------------------------------------------------------------------------------
-- proficience
------------------------------------------------------------------------------------
local BI = LibStub( 'LibBabble-Inventory-3.0'):GetLookupTable()
local NOTHING = {}
local PROFICIENCY = {
	[BI['Idols']]   = { DEATHKNIGHT = true,               HUNTER = true, MAGE = true, PALADIN = true, PRIEST = true, ROGUE = true, SHAMAN = true, WARLOCK = true, WARRIOR = true },
	[BI['Librams']] = { DEATHKNIGHT = true, DRUID = true, HUNTER = true, MAGE = true,                 PRIEST = true, ROGUE = true, SHAMAN = true, WARLOCK = true, WARRIOR = true },
	[BI['Leather']] = {                                                  MAGE = true,                 PRIEST = true,                              WARLOCK = true },
	[BI['Mail']]    = {                     DRUID = true,                MAGE = true,                 PRIEST = true, ROGUE = true,                WARLOCK = true },
	[BI['Plate']]   = {                     DRUID = true, HUNTER = true, MAGE = true,                 PRIEST = true, ROGUE = true, SHAMAN = true, WARLOCK = true },
	[BI['Shields']] = { DEATHKNIGHT = true, DRUID = true, HUNTER = true, MAGE = true,                 PRIEST = true, ROGUE = true,                WARLOCK = true },
	[BI['Totems']]  = { DEATHKNIGHT = true, DRUID = true, HUNTER = true, MAGE = true, PALADIN = true, PRIEST = true, ROGUE = true,                WARLOCK = true, WARRIOR = true },
	--
	[BI['Bows']]              = { DEATHKNIGHT = true, DRUID = true,                MAGE = true, PALADIN = true, PRIEST = true,               SHAMAN = true, WARLOCK = true },
	[BI['Crossbows']]         = { DEATHKNIGHT = true, DRUID = true,                MAGE = true, PALADIN = true, PRIEST = true,               SHAMAN = true, WARLOCK = true },
	[BI['Daggers']]           = { DEATHKNIGHT = true,                                           PALADIN = true },
	[BI['Fist Weapons']]      = { DEATHKNIGHT = true,                              MAGE = true, PALADIN = true, PRIEST = true,                              WARLOCK = true },
	[BI['Guns']]              = { DEATHKNIGHT = true, DRUID = true,                MAGE = true, PALADIN = true, PRIEST = true,               SHAMAN = true, WARLOCK = true },
	[BI['One-Handed Axes']]   = {                     DRUID = true,                MAGE = true,                 PRIEST = true,                              WARLOCK = true },
	[BI['One-Handed Maces']]  = {                                   HUNTER = true, MAGE = true,                                                             WARLOCK = true },
	[BI['One-Handed Swords']] = {                     DRUID = true,                                             PRIEST = true,               SHAMAN = true },
	[BI['Polearms']]          = {                                                  MAGE = true,                 PRIEST = true, ROGUE = true, SHAMAN = true, WARLOCK = true },
	[BI['Staves']]            = { DEATHKNIGHT = true,                                           PALADIN = true,                ROGUE = true },
	[BI['Thrown']]            = { DEATHKNIGHT = true, DRUID = true,                MAGE = true, PALADIN = true, PRIEST = true,               SHAMAN = true, WARLOCK = true },
	[BI['Two-Handed Axes']]   = {                     DRUID = true,                MAGE = true,                 PRIEST = true, ROGUE = true,                WARLOCK = true },
	[BI['Two-Handed Maces']]  = {                                   HUNTER = true, MAGE = true,                 PRIEST = true, ROGUE = true,                WARLOCK = true },
	[BI['Two-Handed Swords']] = {                     DRUID = true,                MAGE = true,                 PRIEST = true, ROGUE = true, SHAMAN = true, WARLOCK = true },
	[BI['Wands']]             = { DEATHKNIGHT = true, DRUID = true, HUNTER = true,              PALADIN = true,                ROGUE = true, SHAMAN = true, WARRIOR = true },
	noOffhand                 = {                     DRUID = true,                MAGE = true, PALADIN = true, PRIEST = true,                              WARLOCK = true },
}
local LOOKUP = {
	['Idols']   = 23198,
	['Librams'] = 23203,
	['Leather'] = 6171,
	['Mail']    = 2385,
	['Plate']   = 14896,
	['Shields'] = 17184,
	['Totems']  = 23199,
	--
	['Bows']              = 2504,
	['Crossbows']         = 15807,
	['Daggers']           = 2494,
	['Fist Weapons']      = 15905,
	['Guns']              = 2509,
	['One-Handed Axes']   = 2490,
	['One-Handed Maces']  = 2524,
	['One-Handed Swords'] = 2131,
	['Polearms']          = 15810,
	['Staves']            = 2495,
	['Thrown']            = 29007,
	['Two-Handed Axes']   = 2491,
	['Two-Handed Maces']  = 2493,
	['Two-Handed Swords'] = 2489,
	['Wands']             = 5208,
}

local function GetLookupTable()
	local tab = {}
	for k,v in pairs( LOOKUP) do
		local iName, iLink, iRarity, iLevel, iMinLevel, iType, iSubType = GetItemInfo( link)
		if iSubType then
			tab[k] = iSubType
			LOOKUP[k] = nil
		end
	end
	return tab
end

------------------------------------------------------------------------------------
-- local
------------------------------------------------------------------------------------
local function GetCache( compiled)
	return cache[compiled]
end

local function SetCache( compiled, fkt)
	cache[compiled] = fkt
end

local function GetQuality( rarity)
	local q = qualities[rarity]
	if not q then
		q = getglobal( 'ITEM_QUALITY'..rarity..'_DESC')
		qualities[rarity] = q
	end
	return q
end

local function GetItemID( link)
	local id = link and link:match( '|Hitem:(%d+)') -- '|c%s|Hitem:%i:0:0:0:0:0:0:0|h[%s]|h|r'
	return tonumber( id)
end

local function GetItemValue( info, link)
	local iName, iLink, iRarity, iLevel, iMinLevel, iType, iSubType, iStackCount, iEquipLoc, iTexture, iPrice = GetItemInfo( link)
	if info == 'name' then
		return iName
	end
	if info == 'id' then
		return GetItemID( iLink) or 0
	end
	if info == 'quality' then
		return GetQuality( iRarity or 0)
	end
	if info == 'level' then
		return tonumber( iLevel) or 0
	end
	if info == 'uselevel' then
		return tonumber( iMinLevel) or 0
	end
	if info == 'type' then
		return iType
	end
	if info == 'subtype' then
		return iSubType
	end
	if info == 'count' then
		return tonumber( iStackCount) or 1
	end
	if info == 'equip' then
		return iEquipLoc and _G[iEquipLoc]
	end
	if info == 'price' then
		return tonumber( iPrice) or 0
	end
	return ''
end

------------------------------------------------------------------------------------
-- environment
------------------------------------------------------------------------------------
local Environment = {
	['name'] = function( link, ...)
		local iName = GetItemValue( 'name', link)
		for n = 1,select( '#', ...) do
			local name = select( n, ...)
			if iName == name then
				return true
			end
		end
		return false
	end,
	
	['id'] = function( link, ...)
		local iID = GetItemValue( 'id', link)
		for n = 1,select( '#', ...) do
			local value = select( n, ...)
			local id = tonumber( value) or 0
			if iID == id then
				return true
			end
		end
		return false
	end,
	
	['quality'] = function( link, ...)
		local iQuality = GetItemValue( 'quality', link)
		for n = 1,select( '#', ...) do
			local quality = select( n, ...)
			if iQuality == quality then
				return true
			end
		end
		return false
	end,
	
	['level'] = function( link, min, max)
		local iLevel = GetItemValue( 'level', link)
		local iMin = tonumber( min) or 0
		local iMax = tonumber( max)
		if iMax then
			return iMax >= iLevel and iLevel >= iMin
		else
			return iLevel >= iMin
		end
	end,
	
	['uselevel'] = function( link, min)
		local iLevel = GetItemValue( 'uselevel', link)
		local iMin = tonumber( min) or 0
		return iLevel >= iMin
	end,
	
	['type'] = function( link, ...)
		local iType = GetItemValue( 'type', link)
		for n = 1,select( '#', ...) do
			local type = select( n, ...)
			if iType == type then
				return true
			end
		end
		return false
	end,
	
	['subtype'] = function( link, ...)
		local iSubType = GetItemValue( 'subtype', link)
		for n = 1,select( '#', ...) do
			local subtype = select( n, ...)
			if iSubType == subtype then
				return true
			end
		end
		return false
	end,
	
	['count'] = function( link, min)
		local iCount = GetItemValue( 'count', link)
		local iMin = tonumber( min) or 1
		return iCount >= iMin
	end,
	
	['equip'] = function( link, ...)
		local iEquip = GetItemValue( 'equip', link)
		for n = 1,select( '#', ...) do
			local equip = select( n, ...)
			if iEquip == equip then
				return true
			end
		end
		return false
	end,

	['useable'] = function( link, ...)
	local iName, iLink, iRarity, iLevel, iMinLevel, iType, iSubType, iStackCount, iEquip = GetItemInfo( link)
		if iType ~= ARMOR and iType ~= ENCHSLOT_WEAPON then
			return true
		end
		local tbl = iSubType and PROFICIENCY[iSubType]
		if not tbl then
			return true
		end
		local noOffhand = (iEquip == 'INVTYPE_WEAPONOFFHAND') and (iType == ENCHSLOT_WEAPON) and PROFICIENCY.noOffhand or NOTHING
		local cnt = select( '#', ...)
		if cnt == 0 then
			local _, class = UnitClass( 'player')
			if not (tbl[class] or (noOffhand and noOffhand[class])) then
				return true
			end
		else
			for n = 1,cnt do
				local class = select( n, ...)
				if not (tbl[class] or noOffhand[class]) then
					return true
				end
			end
		end
		return false
	end,

	['price'] = function( link, min, max)
		local iPrice = GetItemValue( 'price', link)
		local iMin = tonumber( min) or 0
		local iMax = tonumber( max)
		if iMax then
			return iMax >= iPrice and iPrice >= iMin
		else
			return iLevel >= iMin
		end
	end,
	
	['bag'] = function( link, ...)
		for n = 1,select( '#', ...) do
			local bagID = tonumber( select( n, ...))
			if bagID and 0 <= bagID and bagID <= NUM_BAG_SLOTS then
				for slotID = 1,GetContainerNumSlots( bagID) do
					local itemLink = GetContainerItemLink( bagID, slotID)
					if link == itemLink then
						return true
					end
				end
			end
		end
		return false
	end,

	['outfit'] = function( link, ...)
		if IsAddOnLoaded( 'Outfitter') and Outfitter and Outfitter:IsInitialized() then
			local info = Outfitter:GetItemInfoFromLink( link )
			local outfits = info and Outfitter:GetOutfitsUsingItem( info )
			if outfits and #outfits > 0 then
				for n = 1,select( '#', ...) do
					local value = select( n, ...)
					for _,outfit in pairs( outfits) do
						if outfit and outfit == value then
							return true
						end
					end
				end
			end
		end
		return false
	end,
	
	['periodictable'] = function( link, ...)
		local PT3 = LibStub( 'LibPeriodicTable-3.1', true)
		if PT3 then
			for n = 1,select( '#', ...) do
				local set = select( n, ...)
				if PT3:ItemInSet( link, set) then
					return true
				end
			end
		end
		return false
	end,
	
	['tooltip'] = function( link, ...)
		local GRA = LibStub( 'LibGratuity-3.0', true)
		if GRA then
			GRA:SetHyperlink( link)
			for n = 1,select( '#', ...) do
				local value = select( n, ...)
				if GRA:Find( value) then
					return true
				end
			end
		end
		return false
	end,
}

------------------------------------------------------------------------------------
-- lib
------------------------------------------------------------------------------------
local function Link( link)
	local id = GetItemID( link)
	if id then
		return 'ref@'..id
	end
	return link
end

local function Trim( s)
	return '"'..s:match( '^[%s"\']*(.-)[%s"\']*$')..'"'
end

local function Function( fkt, params)
	fkt = fkt:lower()
	params = params:gsub( '|c%x+|H[^%]]+%]|h|r', Link) -- '|c%s|Hitem:%i:0:0:0:0:0:0:0|h[%s]|h|r'
	params = params:gsub( 'ref@%d+', function( ref)
		local id = tonumber( ref:sub( 5))
		if id then
			return GetItemValue( fkt, id)
		end
		return ref
	end)
	return fkt..'(link,'..params:gsub( '[^,]+', Trim)
end

function lib:Compile( template)
	local code = tostring( template)
	code = code:gsub( '%s+', ' ')
	code = code:gsub( '(%a+)%(([^%)]*)', Function)
	return 'local link = ... return ' .. code
end

function lib:Execute( compiled, link)
	local fkt = GetCache( compiled)
	if not fkt then
		fkt = assert( loadstring( compiled))
		setfenv( fkt, Environment)
		SetCache( compiled, fkt)
	end
	return fkt( link)
end
